/**
 * Copyright (c) Milad Eftekhar <milad@cs.toronto.edu>
 * @author Milad
 *
 */


import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.PriorityQueue;


public class CGC {
	static int MAXITERATION;
	public static float influentialgroups(int l, int budget, String graphfile, String groupfile, String membershipfile, final float PROB, final int BETA,final int R) throws IOException{
		
		/*
		 * input: 
		 * l = the number of wanted influential groups
		 * budget = the advertising budget (k)
		 * graphfile = the filename containing the social graph.
		 * groupfile = the filename containing groups and their members (group -> a list of users): dblp_MClusters.bin
		 * membershipfile = the filename containing each user is a member of what groups (user -> a list of groups): dblp_groups.bin
		 * 
		 * output:
		 * print: the l most influential groups
		 * return: the average cascade value (the number of convinced nodes) when the seedset is the printed nodes (the l most influential nodes) and the experiment is simulated for R different delta settings.
		 *
		 */
		
		
	

		//printing parameter setting
		System.out.println("l: "+l);
		System.out.println("budget: "+budget);
		System.out.println("graphfile: "+graphfile);
		System.out.println("groupfile: "+groupfile);
				
		//Reading the social graph
		ArrayList<weightedEdge>[] socialgraph = readFile.readEdges_from_file(graphfile, new Hashtable<HTNode, Integer>());	

		//Reading the MClusters (a group -> a list of users belonging to that group) 
		ArrayList<Integer>[] groups = readFile.readMClusters_from_file(groupfile);
		int g_dimension = groups.length;
		
		//Reading the group memberships ( a user -> a list of groups this user belongs to)
		ArrayList<Integer>[] memberships = readFile.readMClusters_from_file(membershipfile);
		
		long before_modeling = new Date().getTime();
		
		float[][] groupmodel = modeling(socialgraph, groups, memberships);
		float[] rho = cohesionfactors(socialgraph, groups, memberships);
		
		long after_modeling = new Date().getTime();
		System.out.println("The run time of modeling (ms) is: "+(after_modeling-before_modeling));
		
		int[] InfNodes = new int[l];
		
		long time = 0;
		float group_budget = (float) BETA*budget/l;
		
		for (int repeat = 0; repeat < 1000; repeat++){
			//Measuring run time
			long start = new Date().getTime();
	
			
			
			boolean[] used = new boolean[g_dimension];
			InfNodes = new int[l];
			
			
				
			float maxcascade = 0;
	
			for(int i = 0; i<l; i++){
				
	//			System.out.println("Start to explore for "+i+"th influnetial group");
				maxcascade=-1; int maxnode=-1;
				for(int j=0; j<g_dimension; j++){
					if (!used[j]){
						InfNodes[i] = j;
						
						float sum_sim = compute_Revenue(InfNodes,i,group_budget,groups,memberships,groupmodel,rho,PROB);
						
						if (sum_sim>maxcascade){
							maxcascade = sum_sim;
							maxnode = j;
						}
					}
				}
				
				InfNodes[i] = maxnode;
				used[maxnode]=true;
				
	//			System.out.println(i+"th influential group is: "+maxnode);
	//			System.out.println("Cascade is: "+maxcascade);
				
			}
			long finish = new Date().getTime();
			
			time += (finish-start);
			
		}
		
		//Printing the run time
		
		System.out.println("The run time (ms) is: "+((float) time/1000));
		
		//Printing k Influential Nodes
		System.out.println(l+" Influential groups:");
		for(int infnode : InfNodes)
			System.out.print(infnode+" ");
		System.out.println();
		
		
		
		
		
		//returns the average cascade value (the number of convinced nodes) of seedset InfNodes for R different delta settings.
		//This is a simulation of linear threshold models consistent with results of IC and FGC to be comparable.
		
		float cascade = 0;
		for(int phase = 0; phase < R; phase++)
			cascade += FGC.compute_Revenue(InfNodes, l-1, group_budget, socialgraph, groups);		
		return (float) cascade/R;

	}
	

	public static float[][] modeling(ArrayList<weightedEdge>[] socialgraph, ArrayList<Integer>[] groups, ArrayList<Integer>[] memberships){
		int gsize = groups.length;
		float[][] weight = new float[gsize][gsize];
		for(int i=0; i<gsize; i++){
			for(int j=i+1; j<gsize; j++){
				if(i!=j){
					int size1 = groups[i].size();
					int size2 = groups[j].size();
					int minindex = size1<=size2?i:j;
					int maxindex = size1<=size2?j:i;
					float w_ij = 0;
					float w_ji = 0;
					
					int intersection_ij = 0;
					
					for(int member : groups[minindex]){
						boolean in_intersection = false;
						for(int gr1 : memberships[member]){
							if(gr1 == maxindex){
								in_intersection = true;
								++intersection_ij;
								break;
							}
						}
						if(!in_intersection){
							for(weightedEdge neighborobj : socialgraph[member]){
								boolean evaluate = false;
								for(int gr2 : memberships[neighborobj.getNeighbor()]){
									if (gr2 == minindex)	{evaluate = false;	break;}
									if (gr2 == maxindex)	evaluate = true;
									
								}
								if(evaluate){
									if(minindex==i){
										w_ij += neighborobj.getWeight();
										w_ji += neighborobj.getReverseWeight();
									}else{
										w_ji += neighborobj.getWeight();
										w_ij += neighborobj.getReverseWeight();
									}
								}
							}
						}
					}
					
					if(size2!=0)	weight[i][j] = (float) (w_ij+intersection_ij)/size2;
					if(size1!=0)	weight[j][i] = (float) (w_ji+intersection_ij)/size1;
					
				}
			}
		}
		return weight;
	}
	
	public static float[] cohesionfactors(ArrayList<weightedEdge>[] socialgraph, ArrayList<Integer>[] groups, ArrayList<Integer>[] memberships){
		int gsize = groups.length;
		float[] rho = new float[gsize];
		for(int i=0; i<gsize; i++){
			float sum = 0;
			for(Integer member : groups[i]){
				for(weightedEdge neighborobj : socialgraph[member]){
					boolean ismember = false;
					for(int gr : memberships[neighborobj.getNeighbor()]){
						if (gr == i){
							ismember = true;
							break;
						}
					}
					if(ismember){
						sum += neighborobj.getWeight();
					}
				}
			}
			
			rho[i] = (float) sum/groups[i].size();
		}
		return rho;
	}


	public static float compute_Revenue(int[] seedset, int index, float group_budget, ArrayList<Integer>[] groups, ArrayList<Integer>[] memberships, float[][] groupmodel, float[] rho, final float PROB){
		
		int gsize = groupmodel.length;
		float[] NProg = new float[gsize];			// the new progress percentage of each cluster
		float[] OProg = new float[gsize];			// the last progress percentage which each cluster has used to diffuse strategies
		
		PriorityQueue<Node> active = new PriorityQueue<Node>();
		
		for(int i=0; i<=index; i++){
			float denominator = groups[seedset[i]].size()*(1-PROB*rho[seedset[i]]);
			NProg[seedset[i]] =min((float) group_budget/denominator,1);
			active.add(new Node(seedset[i],NProg[seedset[i]]));
		}

			
		int IterationNumber = 0;				
		Node top;
		while((top=active.poll())!=null&&(IterationNumber < MAXITERATION)){	//The iteration number check was at if(NProg[i]>tp) line.
			++IterationNumber;
			int topIndex = top.getind();
			
			for(int i=0; i<gsize; i++){
				if(groupmodel[topIndex][i]>0){
					float tp = NProg[i];
					NProg[i] = min( 1 , NProg[i] + (float) (NProg[topIndex]-OProg[topIndex])*groupmodel[topIndex][i]*PROB*(1-NProg[i])/(1-PROB*rho[i]) );
					
				
				///	Update(active,i,NProg[i])
					if(NProg[i]>tp){ 
						active.remove(new Node(i,tp-OProg[i]));
						active.add(new Node(i, NProg[i]-OProg[i]));
					}
				}
			}
			OProg[topIndex] = NProg[topIndex];
			
			//Calculate this iteration value. For convergence experiment.
			/*
			float cascade = 0;
			for(int i=0; i<gsize; i++){
				cascade += (float) NProg[i]*groups[i].size();
			}
			
			for(ArrayList<Integer> user_groups : memberships){
				if(user_groups.size()==2){
					cascade -= (NProg[user_groups.get(0)]*NProg[user_groups.get(1)]);
				}
				if(user_groups.size()==3){
					cascade -= (NProg[user_groups.get(0)]*NProg[user_groups.get(1)] + NProg[user_groups.get(0)]*NProg[user_groups.get(2)] + NProg[user_groups.get(1)]*NProg[user_groups.get(2)] 
							- NProg[user_groups.get(0)]*NProg[user_groups.get(1)]*NProg[user_groups.get(2)]);
				}
			}
			System.out.println("Iteration: "+IterationNumber+", Cascade: "+cascade);
			*/
		}

		//Calculate final group cascade
		float cascade = 0;
		for(int i=0; i<gsize; i++){
			cascade += (float) NProg[i]*groups[i].size();
		}
		
		for(ArrayList<Integer> user_groups : memberships){
			if(user_groups.size()==2){
				cascade -= (NProg[user_groups.get(0)]*NProg[user_groups.get(1)]);
			}
			if(user_groups.size()==3){
				cascade -= (NProg[user_groups.get(0)]*NProg[user_groups.get(1)] + NProg[user_groups.get(0)]*NProg[user_groups.get(2)] + NProg[user_groups.get(1)]*NProg[user_groups.get(2)] 
						- NProg[user_groups.get(0)]*NProg[user_groups.get(1)]*NProg[user_groups.get(2)]);
			}
		}

		return cascade;
		
	}
	public static float min(float a, float b){
		return a<=b?a:b;
	}
	
	
	public static void main(String[] args) throws IOException{
		//args[0]: l the number of desired groups
		//args[1]: advertising budget (k)
		//args[2]: the filename containing the social graph
		//args[3]: the filename containing groups (group -> a list of users): dblp_MClusters.bin
		//args[4]: the filename containing memberships (user -> a list of groups): dblp_groups.bin
		//args[5]: PROB
		//args[6]: BETA
		//args[7]: R
		
		MAXITERATION = 30;
//		for (MAXITERATION = 5; MAXITERATION<=200; MAXITERATION=MAXITERATION+5){
	//		System.out.println("MAXITERATION: "+ MAXITERATION);
			System.out.println("cascade: "+influentialgroups(Integer.parseInt(args[0]),Integer.parseInt(args[1]),args[2],args[3],args[4],Float.parseFloat(args[5]),Integer.parseInt(args[6]),Integer.parseInt(args[7])));
	//	}
	}
}
