
/**
 * Copyright (c) Milad Eftekhar <milad@cs.toronto.edu>
 * @author Milad
 *
 */


import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.LinkedList;


public class IC {
	//Individual Cascade.
	//A class to determine k influential nodes of a network.
	//The results have been published in KDD 2013.
	
	
	
	public static float influentialnodes(int k, String graphfile, final int R) throws IOException{ 	
		
		/*
		 * input: 
		 * k = the number of wanted influential nodes
		 * graphfile = the filename containing the social graph.
		 * thresholdfile = the filename containing R different threshold settings.
		 * 
		 * output:
		 * print: the k most influential nodes
		 * return: the average cascade value (the number of convinced nodes) when the seedset is the printed nodes (the k most influential nodes) and the experiment is simulated for R different delta settings.
		 *
		 */
		
		//Print parameter setting
		System.out.println("k: "+k);
		System.out.println("graphfile: "+graphfile);
		

		
		//Reading the social graph
		ArrayList<weightedEdge>[] socialgraph = readFile.readEdges_from_file(graphfile, new Hashtable<HTNode, Integer>());	
		int dimension = socialgraph.length;
		System.out.println("dimension is: "+dimension);
		int num_edge = 0;
		for(ArrayList<weightedEdge> user : socialgraph)
			num_edge += user.size();
		System.out.println("Number of directed edges: "+num_edge);
		
		
		//Initialization
		boolean[] used = new boolean[dimension];
		int[] InfNodes = new int[k];
		
		//Measuring run time
		long start = new Date().getTime();
			
		float maxcascade = 0;


		for(int i = 0; i<k; i++){
			
			System.out.println("The run time (ms) for k = "+i+" is: "+(new Date().getTime()-start));

			System.out.println("Start to explore for "+i+"th influnetial node");
			maxcascade=-1; int maxnode=-1;
			for(int j=0; j<dimension; j++){
				if (!used[j]){
					InfNodes[i] = j;
					
					float sum_sim=0;	//saves the sum of val in R simulations (different delta settings).
					for(int phase=0; phase<R; phase++){
						sum_sim += computeFaster(InfNodes,i,socialgraph);
					}
					sum_sim/=R;
					
					if (sum_sim>maxcascade){
						maxcascade = sum_sim;
						maxnode = j;
					}
				}
			}
			
			InfNodes[i] = maxnode;
			used[maxnode]=true;
			
			System.out.println(i+"th influential node is: "+maxnode);
			System.out.println("Cascade is: "+maxcascade);
			
		}
		
		long finish = new Date().getTime();
		
		//Printing the run time
		System.out.println("The run time (ms) is: "+(finish-start));
		
		//Printing k Influential Nodes
		System.out.println(k+" Influential nodes:");
		for(int infnode : InfNodes)
			System.out.print(infnode+" ");
		System.out.println();
		
		//return the average cascade value (the number of convinced nodes) of seedset InfNodes for R different delta settings.
		return maxcascade;		
	}
	
public static int computeFaster(int[] seedset, int index, ArrayList<weightedEdge>[] socialgraph){
		
		/* Applicable when init_inf all members in seedset is 1 that happens when we run the individual case IC. 
		 * When we call IC for other algorithms like FGC and CGC we cannot use this function since init_inf is not 1 (it is a number between 0 and 1); 
		 * hence we need to use compute(...) in these cases.
		 * 
		 * input:
		 * seedset: the seed set
		 * index: the index of last member in seedset = |seedset|-1
		 * init_inf: the influenced value for the seedset members.
		 * socialgraph
		 * delta: showing the threshold for each node in the graph
		 * dimension: size of the social graph
		 * 
		 * output:
		 * return: the cascade (number of convinced nodes) as the result of convincing seed set Nodes. 
		 */
		
		int cascade = 0;
		int dimension = socialgraph.length;
		
		float[] delta = new float[dimension];
		for (int i=0; i<dimension; i++)
			delta[i] = (float) Math.random();
		
		float[] init_inf = new float[dimension];

		//Initializing the active nodes.
		LinkedList<Integer> active = new LinkedList<Integer>();
		for(int i=0; i<= index; i++){
				active.add(seedset[i]);
				init_inf[seedset[i]]=1;
				++cascade;
		}
		
		//Simulation.
		Integer current;
		while((current=active.poll())!=null){
			
			for(weightedEdge edge : socialgraph[current]){
				int neighbor = edge.getNeighbor();
				float newval = init_inf[neighbor]+edge.getWeight();
				if(newval>=delta[neighbor] && init_inf[neighbor]<delta[neighbor]){
					active.add(neighbor);
					init_inf[neighbor]=1;
					++cascade;
				}
				if(newval<delta[neighbor])
					init_inf[neighbor]=newval;
			}
		}
		

		return cascade;
	}

	
	public static int compute(float[] init_inf, ArrayList<weightedEdge>[] socialgraph, float[] delta){
		
		/*
		 * input:
		 * seedset: the seed set
		 * index: the index of last member in seedset = |seedset|-1
		 * init_inf: the influenced value for the seedset members.
		 * socialgraph
		 * delta: showing the threshold for each node in the graph
		 * dimension: size of the social graph
		 * 
		 * output:
		 * return: the cascade (number of convinced nodes) as the result of convincing seed set Nodes. 
		 */
		
		int cascade = 0;
		//Initializing the active nodes.
		int dimension = socialgraph.length;
		LinkedList<Integer> active = new LinkedList<Integer>();
		for(int i=0; i< dimension; i++){
			if(init_inf[i] > delta[i]){						
				active.add(i);
				init_inf[i]=1;
				++cascade;
			}
		}
		
		//Simulation.
		Integer current;
		while((current=active.poll())!=null){
			
			for(weightedEdge edge : socialgraph[current]){
				int neighbor = edge.getNeighbor();
				float newval = init_inf[neighbor]+edge.getWeight();
				if(newval>=delta[neighbor] && init_inf[neighbor]<delta[neighbor]){
					active.add(neighbor);
					init_inf[neighbor]=1;
					++cascade;
				}
				if(newval<delta[neighbor])
					init_inf[neighbor]=newval;
			}
		}
		

		return cascade;
	}
	
	public static void main(String[] args) throws IOException{
		//args[0]: k the number of influential nodes expected
		//args[1]: the filename for social graph
		//args[2]: R
		
		System.out.println(influentialnodes(Integer.parseInt(args[0]),args[1],Integer.parseInt(args[2])));
		
	}

}
